@@ -32,12 +32,12 @@ from firewall.core import ipset
 from firewall.core import modules
 from firewall.core.fw_icmptype import FirewallIcmpType
 from firewall.core.fw_service import FirewallService
-from firewall.core.fw_zone import FirewallZone
-from firewall.core.fw_direct import FirewallDirect
+from firewall.core.fw_zone import FirewallZoneIPTables
+from firewall.core.fw_direct import FirewallDirectIPTables
 from firewall.core.fw_config import FirewallConfig
 from firewall.core.fw_policies import FirewallPolicies
 from firewall.core.fw_ipset import FirewallIPSet
-from firewall.core.fw_transaction import FirewallTransaction, reverse_rule
+from firewall.core.fw_transaction import FirewallTransaction
 from firewall.core.fw_helper import FirewallHelper
 from firewall.core.logger import log
 from firewall.core.io.firewalld_conf import firewalld_conf
@@ -81,8 +81,8 @@ class Firewall(object):
 
         self.icmptype = FirewallIcmpType(self)
         self.service = FirewallService(self)
-        self.zone = FirewallZone(self)
-        self.direct = FirewallDirect(self)
+        self.zone = FirewallZoneIPTables(self)
+        self.direct = FirewallDirectIPTables(self)
         self.config = FirewallConfig(self)
         self.policies = FirewallPolicies()
         self.ipset = FirewallIPSet(self)
@@ -370,8 +370,8 @@ class Firewall(object):
             try:
                 obj.read()
             except Exception as msg:
-                log.debug1("Failed to load direct rules file '%s': %s",
-                           config.FIREWALLD_DIRECT, msg)
+                log.error("Failed to load direct rules file '%s': %s",
+                          config.FIREWALLD_DIRECT, msg)
         self.direct.set_permanent_config(obj)
         self.config.set_direct(copy.deepcopy(obj))
 
@@ -391,14 +391,11 @@ class Firewall(object):
         # Start transaction
         transaction = FirewallTransaction(self)
 
-        if reload:
-            self.set_policy("DROP", use_transaction=transaction)
-
         # flush rules
         self.flush(use_transaction=transaction)
 
         # If modules need to be unloaded in complete reload or if there are
-        # ipsets to get applied, limit the transaction to set_policy and flush.
+        # ipsets to get applied, limit the transaction to flush.
         #
         # Future optimization for the ipset case in reload: The transaction
         # only needs to be split here if there are conflicting ipset types in
@@ -440,14 +437,18 @@ class Firewall(object):
 
         # apply direct chains, rules and passthrough rules
         if self.direct.has_configuration():
-            transaction.enable_generous_mode()
             log.debug1("Applying direct chains rules and passthrough rules")
             self.direct.apply_direct(transaction)
 
-            # Execute transaction
-            transaction.execute(True)
-            transaction.disable_generous_mode()
-            transaction.clear()
+            # since direct rules are easy to make syntax errors lets highlight
+            # the cause if the transaction fails.
+            try:
+                transaction.execute(True)
+                transaction.clear()
+            except FirewallError as e:
+                raise FirewallError(e.code, "Direct: %s" % (e.msg if e.msg else ""))
+            except Exception:
+                raise
 
         del transaction
 
@@ -456,11 +457,16 @@ class Firewall(object):
             tm2 = time.time()
             log.debug2("Flushing and applying took %f seconds" % (tm2 - tm1))
 
-        self._state = "RUNNING"
-
     def start(self):
-        self._start()
-        self.set_policy("ACCEPT")
+        try:
+            self._start()
+        except Exception:
+            self._state = "FAILED"
+            self.set_policy("ACCEPT")
+            raise
+        else:
+            self._state = "RUNNING"
+            self.set_policy("ACCEPT")
 
     def _loader(self, path, reader_type, combine=False):
         # combine: several zone files are getting combined into one obj
@@ -520,7 +526,7 @@ class Firewall(object):
                     # add a deep copy to the configuration interface
                     self.config.add_service(copy.deepcopy(obj))
                 elif reader_type == "zone":
-                    obj = zone_reader(filename, path)
+                    obj = zone_reader(filename, path, no_check_name=combine)
                     if combine:
                         # Change name for permanent configuration
                         obj.name = "%s/%s" % (
@@ -671,35 +677,13 @@ class Firewall(object):
 
     # apply default rules
     def __apply_default_rules(self, ipv, transaction):
-        default_rules = { }
-
-        if ipv in [ "ipv4", "ipv6" ]:
-            x = ipXtables
+        if ipv is "ipv4":
+            x = self.ip4tables_backend
+        elif ipv is "ipv6":
+            x = self.ip6tables_backend
         else:
-            x = ebtables
-        for table in x.DEFAULT_RULES:
-            default_rules[table] = x.DEFAULT_RULES[table][:]
-
-        if self._log_denied != "off":
-            for table in x.LOG_RULES:
-                default_rules.setdefault(table, []).extend(x.LOG_RULES[table])
-
-        for table in default_rules:
-            if table not in self.get_available_tables(ipv):
-                continue
-            prefix = [ "-t", table ]
-            for rule in default_rules[table]:
-                if type(rule) == list:
-                    _rule = prefix + rule
-                else:
-                    _rule = prefix + functions.splitArgs(rule)
-                #if self._individual_calls or \
-                #   (ipv == "eb" and not
-                #    self.ebtables_backend.restore_noflush_option):
-                #    self.rule(ipv, _rule)
-                #else:
-                #    transaction.add_rule(ipv, _rule)
-                transaction.add_rule(ipv, _rule)
+            x = self.ebtables_backend
+        x.apply_default_rules(transaction, self._log_denied)
 
     def apply_default_rules(self, use_transaction=None):
         if use_transaction is None:
@@ -718,22 +702,7 @@ class Firewall(object):
             # Start new transaction
             transaction.clear()
 
-            # here is no check for ebtables.restore_noflush_option needed
-            # as ebtables is not used in here
-            transaction.add_rule("ipv6",
-                                 [ "-I", "PREROUTING", "1", "-t", "raw",
-                                   "-p", "ipv6-icmp",
-                                   "--icmpv6-type=router-advertisement",
-                                   "-j", "ACCEPT" ]) # RHBZ#1058505
-            transaction.add_rule("ipv6",
-                                 [ "-I", "PREROUTING", "2", "-t", "raw",
-                                   "-m", "rpfilter", "--invert", "-j", "DROP" ])
-            if self._log_denied != "off":
-                transaction.add_rule("ipv6",
-                                     [ "-I", "PREROUTING", "2", "-t", "raw",
-                                       "-m", "rpfilter", "--invert",
-                                       "-j", "LOG",
-                                       "--log-prefix", "rpfilter_DROP: " ])
+            self.ip6tables_backend.apply_rpfilter_rules(transaction, self._log_denied)
 
             # Execute ipv6_rpfilter transaction, it might fail
             try:
@@ -807,48 +776,6 @@ class Firewall(object):
     # rule function used in handle_ functions
 
     def rule(self, ipv, rule):
-        # replace %%REJECT%%
-        try:
-            i = rule.index("%%REJECT%%")
-        except ValueError:
-            pass
-        else:
-            if ipv in [ "ipv4", "ipv6" ]:
-                rule[i:i+1] = [ "REJECT", "--reject-with",
-                                ipXtables.DEFAULT_REJECT_TYPE[ipv] ]
-            else:
-                raise FirewallError(errors.EBTABLES_NO_REJECT,
-                                    "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-
-        # replace %%ICMP%%
-        try:
-            i = rule.index("%%ICMP%%")
-        except ValueError:
-            pass
-        else:
-            if ipv in [ "ipv4", "ipv6" ]:
-                rule[i] = ipXtables.ICMP[ipv]
-            else:
-                raise FirewallError(errors.INVALID_IPV,
-                                    "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-
-        # replace %%LOGTYPE%%
-        try:
-            i = rule.index("%%LOGTYPE%%")
-        except ValueError:
-            pass
-        else:
-            if self._log_denied == "off":
-                return ""
-            if ipv not in [ "ipv4", "ipv6" ]:
-                raise FirewallError(errors.INVALID_IPV,
-                                    "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-            if self._log_denied in [ "unicast", "broadcast", "multicast" ]:
-                rule[i:i+1] = [ "-m", "pkttype", "--pkt-type",
-                                self._log_denied ]
-            else:
-                rule.pop(i)
-
         # remove leading and trailing '"' for use with execve
         i = 0
         while i < len(rule):
@@ -860,15 +787,15 @@ class Firewall(object):
         if ipv == "ipv4":
             # do not call if disabled
             if self.ip4tables_enabled:
-                return self.ip4tables_backend.set_rule(rule)
+                return self.ip4tables_backend.set_rule(rule, self._log_denied)
         elif ipv == "ipv6":
             # do not call if disabled
             if self.ip6tables_enabled:
-                return self.ip6tables_backend.set_rule(rule)
+                return self.ip6tables_backend.set_rule(rule, self._log_denied)
         elif ipv == "eb":
             # do not call if disabled
             if self.ebtables_enabled:
-                return self.ebtables_backend.set_rule(rule)
+                return self.ebtables_backend.set_rule(rule, self._log_denied)
         else:
             raise FirewallError(errors.INVALID_IPV,
                                 "'%s' not in {'ipv4'|'ipv6'|'eb'}" % ipv)
@@ -876,53 +803,7 @@ class Firewall(object):
         return ""
 
     def rules(self, ipv, rules):
-        _rules = [ ]
-
-        for rule in rules:
-            # replace %%REJECT%%
-            try:
-                i = rule.index("%%REJECT%%")
-            except ValueError:
-                pass
-            else:
-                if ipv in [ "ipv4", "ipv6" ]:
-                    rule[i:i+1] = [ "REJECT", "--reject-with",
-                                    ipXtables.DEFAULT_REJECT_TYPE[ipv] ]
-                else:
-                    raise FirewallError(errors.EBTABLES_NO_REJECT,
-                                        "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-
-            # replace %%ICMP%%
-            try:
-                i = rule.index("%%ICMP%%")
-            except ValueError:
-                pass
-            else:
-                if ipv in [ "ipv4", "ipv6" ]:
-                    rule[i] = ipXtables.ICMP[ipv]
-                else:
-                    raise FirewallError(errors.INVALID_IPV,
-                                        "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-
-            # replace %%LOGTYPE%%
-            try:
-                i = rule.index("%%LOGTYPE%%")
-            except ValueError:
-                pass
-            else:
-                if self._log_denied == "off":
-                    continue
-                if ipv not in [ "ipv4", "ipv6" ]:
-                    raise FirewallError(errors.INVALID_IPV,
-                                        "'%s' not in {'ipv4'|'ipv6'}" % ipv)
-                if self._log_denied in [ "unicast", "broadcast",
-                                         "multicast" ]:
-                    rule[i:i+1] = [ "-m", "pkttype", "--pkt-type",
-                                    self._log_denied ]
-                else:
-                    rule.pop(i)
-
-            _rules.append(rule)
+        _rules = rules[:]
 
         backend = None
         if ipv == "ipv4":
@@ -963,14 +844,14 @@ class Firewall(object):
                     log.error(msg)
                     for rule in reversed(_rules[:i]):
                         try:
-                            backend.set_rule(reverse_rule(rule))
+                            backend.set_rule(ipXtables.reverse_rule(rule))
                         except Exception:
                             # ignore errors here
                             pass
                     return False
             return True
         else:
-            return backend.set_rules(_rules)
+            return backend.set_rules(_rules, log_denied=self._log_denied)
 
     # check functions
 
@@ -1046,8 +927,15 @@ class Firewall(object):
         # stop
         self.cleanup()
 
-        # start
-        self._start(reload=True, complete_reload=stop)
+        self.set_policy("DROP")
+
+        start_exception = None
+        try:
+            self._start(reload=True, complete_reload=stop)
+        except Exception as e:
+            # save the exception for later, but continue restoring interfaces,
+            # etc. We'll re-raise it at the end.
+            start_exception = e
 
         # handle interfaces in the default zone and move them to the new
         # default zone if it changed
@@ -1077,8 +965,7 @@ class Firewall(object):
             else:
                 log.info1("New zone '%s'.", zone)
         if len(_zone_interfaces) > 0:
-            keys = _zone_interfaces.keys()
-            for zone in keys:
+            for zone in list(_zone_interfaces.keys()):
                 log.info1("Lost zone '%s', zone interfaces dropped.", zone)
                 del _zone_interfaces[zone]
         del _zone_interfaces
@@ -1093,6 +980,12 @@ class Firewall(object):
         else:
             self.set_policy("ACCEPT")
 
+        if start_exception:
+            self._state = "FAILED"
+            raise start_exception
+        else:
+            self._state = "RUNNING"
+
     # STATE
 
     def get_state(self):
@@ -1142,9 +1035,6 @@ class Firewall(object):
             self._log_denied = value
             self._firewalld_conf.set("LogDenied", value)
             self._firewalld_conf.write()
-
-            # now reload the firewall
-            self.reload()
         else:
             raise FirewallError(errors.ALREADY_SET, value)
 
@@ -1163,9 +1053,6 @@ class Firewall(object):
             self._automatic_helpers = value
             self._firewalld_conf.set("AutomaticHelpers", value)
             self._firewalld_conf.write()
-
-            # now reload the firewall
-            self.reload()
         else:
             raise FirewallError(errors.ALREADY_SET, value)
 
